# Duo, A crazy eights game.
# Copyright (C) 2003-2008 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
[TODO: write this}
"""

import logging
import threading
import time

from duo import ai
from duo import game
from duo import protocol
from duo import referee


logger = logging.getLogger('duo.ai.camille')


class AI(ai.AbstractAI):
    def __init__(self):
        ai.AbstractAI.__init__(self)
        logger.debug('AI.__init__')
        self._condition = threading.Condition()
        self._end = False
        self._heap = None
        self._heap_color = None
        self._is_kitty_present = False
        self._stack_card_taken = False

    def stop_game(self):
        self._condition.acquire()
        self._end = True
        self._condition.notify()
        self._condition.release()

    def on_auth(self, payload, emitter, args):
        pass

    def on_hand(self, message, emitter, args):
        self._condition.acquire()
        if self._stack_card_taken:
            if self._heap.type in ['X', 'D']:
                self._stack_card_taken = False
                self.get_player().send_to_referee(protocol.PassMessage())
            else:
                self._condition.notify()
        self._condition.release()

    def on_turn(self, message, emitter, args):
        # Check it's our turn.
        if message == self.get_player().name:
            logger.debug('%s says my turn!' % self.get_player().name)
            self._condition.acquire()
            self._condition.notify()
            self._condition.release()

    def on_heap(self, message, emitter, args):
        (card, color, player) = message
        self._condition.acquire()
        self._heap = card
        self._heap_color = color
        if self._heap.type in ['X', 'D']:
            self._is_kitty_present = True
        self._condition.release()

    def on_deal(self, message, emitter, args):
        self._condition.acquire()
        self._is_kitty_present = False
        self._condition.release()

    def on_error(self, payload, emitter, args):
        if isinstance(payload, referee.InvalidMoveError):
            logger.debug('%s says sorry!' % self.get_player().name)
            self._condition.acquire()
            self._condition.notify()
            self._condition.release()

    def on_winner(self, payload, emitter, args):
        self._condition.acquire()
        self._end = True
        self._condition.notify()
        self._condition.release()

    def run(self):
        """TODO"""
        logger.debug('%s is back at the table.' % self.get_player().name)
        self._condition.acquire()
        while True:
            # Wait for our turn.
            self._condition.wait()

            # Issue a move.
            if self._end:
                break
            else:
                self._play()
        logger.debug('%s asks for a drink.' % self.get_player().name)
        self._condition.release()

        # Reset between rounds.
        self._end = False
        self._heap = None
        self._heap_color = None
        self._is_kitty_present = False

    def _play(self):
        """TODO"""
        logger.debug('%s is thinking...' % self.get_player().name)
        card_to_play = None
        color_to_play = None

        time.sleep(1)

        hand = self.get_player().hand
        hand_types = [card.type for card in hand if card.type not in ['W','X']]
        hand_colors = [card.color for card in hand if card.color != 'N']
        card_types = list(game.Card.TYPES)
        card_types.reverse()

        # Stack on dangerous cards
        if self._is_kitty_present:
            card_to_play, color_to_play = self._stack_on_kitty()
            if card_to_play:
                logger.debug('%s stacks a %s' % (
                        self.get_player().name, str(card_to_play)))

        # Play by card type (excluding jokers)
        elif self._heap.type in hand_types:
            possibilities = [card for card in hand if card.type == self._heap.type]
            if possibilities:
                card_to_play = possibilities[0]
                color_to_play = card_to_play.color
                logger.debug('%s camille plays %s by card type' % (
                        self.get_player().name, str(card_to_play)))

        # Play by card color
        elif self._heap_color in hand_colors:
            possibilities = [card for card in hand if card.color == self._heap_color]
            if possibilities:
                card_to_play = possibilities[0]
                color_to_play = card_to_play.color
                logger.debug('%s camille plays %s by card color' % (
                        self.get_player().name, str(card_to_play)))

        # Play jokers
        elif 'W' in [card.type for card in hand]:
            card_to_play = game.Card('W', 'N')
            color_to_play = self._choose_color()
            logger.debug('%s plays wild' % self.get_player().name)

        # Play really hard
        elif 'X' in [card.type for card in hand]:
            card_to_play = game.Card('X', 'N')
            color_to_play = self._choose_color()
            logger.debug('%s plays hard!' % self.get_player().name)

        # Play or Draw...
        if card_to_play != None:
            self._stack_card_taken = False
            self.get_player().send_to_referee(
                protocol.PlayMessage(card_to_play, color_to_play))
        elif not self._stack_card_taken:
            logger.debug('%s draws' % self.get_player().name)
            self._stack_card_taken = True
            self.get_player().send_to_referee(protocol.DrawMessage())
        else:
            logger.debug('%s passes' % self.get_player().name)
            self._stack_card_taken = False
            self.get_player().send_to_referee(protocol.PassMessage())


    def _stack_on_kitty(self):
        """TODO"""
        card = None
        color = None
        hand = self.get_player().hand
        cards = filter(
            lambda x: x.type == self._heap.type, self.get_player().hand)
        if len(cards):
            card = cards[0]
            color = card.color
            if color == 'N':
                color = self._choose_color()
        return card, color

    def _choose_color(self):
        """choose the color to ask"""
        chosen_color = 'B'
        max = 0
        colors = [card.color for card in self.get_player().hand
                  if card.color !='N']
        for color in game.Card.COLORS.keys():
            count = colors.count(color)
            if count > max:
                max = count
                chosen_color = color
        logger.debug('%s chooses %s' % (self.get_player().name, chosen_color))
        return chosen_color
